from __future__ import unicode_literals, division, print_function, absolute_import
from builtins import object
import re
import unicodedata
from math import log
from collections import Counter
from scipy.sparse import csr_matrix, dok_matrix
from .dict_utils import norm_dict, invert_dict1, invert_dict2, select_copy


def norm_whitespace(text):
    text = re.sub(r"\s+", " ", text)
    return text.strip()


def preprocess_text(text, to_lower=True, norm_num=True):
    # clean the text: no fucked up characters, html, ...
    nfkd_form = unicodedata.normalize("NFKD", text)
    text = nfkd_form.encode("ASCII", "ignore").decode("ASCII")
    text = re.sub(r"http(s)?://\S*", " ", text)  # remove links (other html crap is assumed to be removed by bs)
    if to_lower:
        text = text.lower()
    if norm_num:
        text = re.sub(r"[0-9]", "1", text)  # normalize numbers
    # clean out non-alphabet characters and normalize whitespace
    text = re.sub(r"[^A-Za-z0-9-]+", " ", text)
    text = re.sub(r"\s+", " ", text)
    return text.strip()


def get_bigram_scores(text_words, min_bgfreq=2.):
    """
    compute scores for identifying bigrams in a collection of texts
    compute unigram and bigram frequencies (of words) and give a score for every bigram.
    depending on this score (e.g. if higher than a threshold), bigram phrases can be identified
    in the raw texts before splitting them into words
    --> once the lists of words in the text_words are replaced by appropriate bigrams, the whole thing could be repeated
        to find trigrams, etc.
    Input:
        - text_words: a list (or generator) of lists of (preprocessed) words
        - min_bgfreq: how often a bigram has to have to occur in the corpus to be recognized (if both words occur just once
                      and that in combination, it still doesn't mean it's a true phrase because we have too little observations of the words
    Returns:
        - bigram_scores: a dict with bigrams ("word1 word2") and their score
    """
    unigram_freq = {}
    bigram_freq = {}
    for wordlist in text_words:
        for i, word in enumerate(wordlist):
            # count unigrams and bigrams
            try:
                unigram_freq[word] += 1.
            except KeyError:
                unigram_freq[word] = 1.
            if i:
                try:
                    bigram_freq["%s %s" % (wordlist[i - 1], word)] += 1.
                except KeyError:
                    bigram_freq["%s %s" % (wordlist[i - 1], word)] = 1.
    # compute bigram scores
    bigram_scores = {}
    for bigram in bigram_freq:
        # discount to ensure a word combination occurred a sufficient amount of times
        if max(0., bigram_freq[bigram] - min_bgfreq):
            bigram_scores[bigram] = bigram_freq[bigram] / \
                max(unigram_freq[bigram.split()[0]], unigram_freq[bigram.split()[1]])
    return bigram_scores


def find_bigrams(textdict, threshold=0.1):
    """
    find bigrams in the texts
    Input:
        - textdict: a dict with {docid: preprocessed_text}
        - threshold: for bigrams scores
    Returns:
        - bigrams: a list of "word1 word2" bigrams
    """
    docids = set(textdict.keys())
    # to identify bigrams, transform the texts into lists of words (assume texts are preprocessed)
    text_words = [textdict[did].split() for did in docids]
    bigram_scores = get_bigram_scores(text_words)
    return [bigram for bigram in bigram_scores if bigram_scores[bigram] > threshold]


def replace_bigrams(textdict, bigrams):
    """
    replace bigrams in the texts
    Input:
        - textdict: a dict with {docid: preprocessed_text}
        - bigrams: for bigrams scores
    Returns:
        - textdict: the same texts but preprocessed and with all bigrams joined as "word1_word2"
    """
    docids = set(textdict.keys())
    # substitute the bigrams in the texts - actually creates trigrams as well :)
    for did in docids:
        text = textdict[did]
        for bigram in bigrams:
            if bigram in text:
                text = text.replace(bigram, "%s_%s" % tuple(bigram.split()))
        textdict[did] = text
    return textdict


def compute_idf(docfeats):
    """
    Inputs:
        - docfeats: a dict with doc_id:{term:count}
    Returns:
        - Dw: a dict with {term: weight}
    """
    # total number of documents
    N = float(len(docfeats))
    # invert the dictionary to be term:{doc_id:count}
    termdocs = invert_dict2(docfeats)
    # compute idf for every term
    return norm_dict({term: log(N / len(termdocs[term])) for term in termdocs})


class FeatureTransform(object):
    """
    FeatureTransform

    a class to transform text into features (similar to sklearn classes)

    Usage:
        # initialize the FeatureTransformer
        ft = FeatureTransform(norm='max', weight=True, renorm='length', identify_bigrams=True)
        # use the training ids to compute the weights but transform all documents
        docfeats = ft.texts2features(textdict, trainids)
        # transform a set of new documents as well using the same weights and identified bigrams
        newdocfeats = ft.texts2features(newtextdict)

    Attributes:
        - to_lower: whether the text should be lower cased in preprocessing
        - norm_num: whether numbers should be normalized, e.g. both years '1989' and '2016' are transformed to '1111'
        - identify_bigrams: if bigrams should be found and replaced
        - norm (binary, max, length, sum, None): how the term counts for each doc should be normalized
        - weight: if idf term weights should be applied
        - renorm: how the features with applied weights should be renormalized
    """

    def __init__(self, norm="max", weight=True, renorm="length", identify_bigrams=True,
                 to_lower=True, norm_num=True, bg_threshold=0.1):
        self.norm = norm
        self.weight = weight
        self.renorm = renorm
        self.identify_bigrams = identify_bigrams
        self.to_lower = to_lower
        self.norm_num = norm_num
        self.bg_threshold = bg_threshold
        self.Dw = {}
        self.bigrams = []

    def texts2features(self, textdict, fit_ids=[]):
        """
        preprocess texts, count how often each word occurs, weight counts, normalize
        If this is called the first time, possibly the idf weights and bigrams are computed (using the documents
            specified in fit_ids), in future calls, the precomputed weights and bigrams are used, e.g. when applying
            the routine to new test documents.
        Input:
            - textdict: a dict with {docid: text}
            - fit_ids: if only a portion of all texts should be used to compute the weights and identify bigrams
                       (e.g. only training data - only used in the first initializing run)
        Returns:
            - docfeats: a dict with {docid: {term: (normalized/weighted) count}}
        """
        docids = set(textdict.keys())
        if not len(fit_ids):
            fit_ids = set(textdict.keys())
        # pre-process texts
        textdict_pp = {did: preprocess_text(textdict[did], self.to_lower, self.norm_num) for did in docids}
        # possibly find bigrams
        if self.identify_bigrams:
            if not self.bigrams:
                self.bigrams = find_bigrams(select_copy(textdict_pp, fit_ids), self.bg_threshold)
            textdict_pp = replace_bigrams(textdict_pp, self.bigrams)
        # split texts into tokens
        docfeats = {}
        for did in docids:
            featdict = dict(Counter(textdict_pp[did].split()))
            # normalize
            if self.norm:
                featdict = norm_dict(featdict, norm=self.norm)
            docfeats[did] = featdict
        # possibly compute idf weights and re-normalize
        if self.weight:
            if not self.Dw:
                self.Dw = compute_idf(select_copy(docfeats, fit_ids))
            for did in docids:
                # if the word was not in Dw (= not in the training set), delete it
                # (otherwise it can mess with renormalization)
                docfeats[did] = {term: docfeats[did][term] * self.Dw[term] for term in docfeats[did] if term in self.Dw}
        if self.renorm:
            for did in docids:
                docfeats[did] = norm_dict(docfeats[did], norm=self.renorm)
        return docfeats


def features2mat(docfeats, docids, featurenames=[]):
    """
    Transform a dictionary with features into a sparse matrix (e.g. for sklearn algorithms)
    Input:
        - docfeats: a dictionary with {docid:{word:count}}
        - docids: the subset of the docfeats (keys to the dict) that should be regarded,
                   defines rows of the feature matrix
        - featurenames: a list of words that define the columns of the feature matrix
                --> when the docids are the training samples, this can be left empty
                    and the words in the training document will be used; for testing give the returned names
    Returns:
        - featmat: a sparse matrix with docids x featurenames
        - featurenames: the list of words defining the columns of the featmat

    Example usage:
        features_train, featurenames = make_featmat(docfeats, trainids)
        features_test, featurenames = make_featmat(docfeats, testids, featurenames)
    """
    if not featurenames:
        featurenames = sorted(invert_dict1(select_copy(docfeats, docids)).keys())
    fnamedict = {feat: i for i, feat in enumerate(featurenames)}
    featmat = dok_matrix((len(docids), len(featurenames)), dtype=float)
    for i, did in enumerate(docids):
        for word in docfeats[did]:
            try:
                featmat[i, fnamedict[word]] = docfeats[did][word]
            except KeyError:
                pass
    featmat = csr_matrix(featmat)
    return featmat, featurenames



from __future__ import unicode_literals, division, print_function, absolute_import
from builtins import range
import numpy as np
from .features import features2mat
from .simcoefs import compute_sim


def dist2kernel(D, perp=30, tol=10e-4, verbose=False):
    """
    Given a distance matrix, select an appropriate sigma for every datapoint such that the perplexity of
    the resulting distribution is equal to perp (see tsne paper)
    Input:
        - D: NxN matrix with distances for all N points to each other
        - perp: desired perplexity (~ # of nearest neighbors)
    Returns:
        - joint probabilities
    """
    logPerp = np.log2(perp)
    # initialize sigmas
    sigmas = np.ones(D.shape[0])
    sigmas_max = 250 * np.ones(D.shape[0])  # unrealistically high value
    sigmas_min = np.zeros(D.shape[0])
    for it in range(50):
        # compute P using the individual sigmas
        P = np.exp(-D * sigmas)
        np.fill_diagonal(P, 0.)    # set diagonal to 0
        # assume the values for each data point are in a column
        P /= np.sum(P, 0)          # normalize
        # compute the entropy of every distribution and how far we're off the desired entropy
        hdiffs = -np.sum(P * np.log2(np.maximum(P, 1e-12)), 0) - logPerp
        err = np.sum(np.abs(hdiffs))
        if verbose:
            print("iteration %i: sum of errors: %.2f" % (it, err))
        if err <= tol:
            break
        # update sigmas
        sigmas_min[hdiffs >= 0] = sigmas[hdiffs >= 0]
        sigmas_max[hdiffs <= 0] = sigmas[hdiffs <= 0]
        sigmas[hdiffs > 0] = 0.5 * (sigmas[hdiffs > 0] + sigmas_max[hdiffs > 0])
        sigmas[hdiffs < 0] = 0.5 * (sigmas[hdiffs < 0] + sigmas_min[hdiffs < 0])
    # sanity check
    if verbose:
        print(np.histogram(sigmas, 20))
    # compute P for real
    P = np.exp(-D * sigmas)
    np.fill_diagonal(P, 0.)     # set diagonal to 0
    P /= np.sum(P, 0)           # normalize
    # and from the conditional probabilities, compute the joint probabilities
    return (P + P.T) / (2. * D.shape[0])


def compute_K(docids, docfeats, sim='linear', normalize=False):
    """
    Imput:
        - docids: list of document ids (keys of docfeats)
        - docfeats: dict with doc_id:{feat:count}
        - sim: type of similarity
        - normalize: if the matrix should be normalized to have 1 on the diagonals and 1 <= s <= 0 on off diagonal
                     by dividing k(x,y) by the mean of k(x,x) and k(y,y)
    Returns:
        - symmetric similarity matrix of size len(docids)xlen(docids)
    """
    # if linear similarity or variant thereof, we're quicker with a dot product
    if sim in ("linear", "cosine", "angularsim", "angulardist"):
        # transform features into matrix
        if isinstance(docfeats, dict):
            featmat, _ = features2mat(docfeats, docids, featurenames=[])
        else:
            featmat = docfeats
        # possibly normalize vectors - doesn't work yet because the featmat is sparse and numpy stupid
        # if sim in ('cosine', 'angularsim', 'angulardist'):
        #     fnorm = np.linalg.norm(featmat, axis=1)
        #     featmat /= fnorm.reshape(featmat.shape[0],1)
        # compute similarity matrix
        S = featmat.dot(featmat.T).toarray()
        # transform further
        if sim in ("angularsim", "angulardist"):
            # make sure values are in the correct range (roundoff errors)
            S = np.minimum(S, 1)
            S = np.maximum(S, 0)
            if sim == "angularsim":
                S = 1. - 2. * np.arccos(S) / np.pi
            else:
                # distance
                S = 2. * np.arccos(S) / np.pi
    else:
        # compute general similarity matrix
        N = len(docids)
        S = np.zeros((N, N))
        for i, did in enumerate(docids):
            for j in range(i + 1):
                similarity = compute_sim(docfeats[did], docfeats[docids[j]], sim)
                S[i, j], S[j, i] = similarity, similarity
    if normalize:
        diag_elem = np.array([np.diag(S)])
        norm = (np.tile(diag_elem, (N, 1)) + np.tile(diag_elem.T, (1, N))) / 2.
        S /= norm
    return S


def compute_K_map(train_ids, test_ids, docfeats, sim="linear", train_idx=[], normalize=False):
    """
    Imput:
        - train_ids, test_ids: list of document ids (keys of docfeats)
        - docfeats: dict with doc_id:feat
        - sim: type of similarity
        - train_idx: list of indexes for train_ids, e.g. for svms we don't need all training examples
    Returns:
        - kernel map of size len(test_ids)xlen(train_ids) with similarities of the test to the training docs
    """
    # if linear similarity or variant thereof, we're quicker with a dot product
    if sim in ("linear", "cosine", "angularsim", "angulardist"):
        # transform features into matrix
        featmat_train, featurenames = features2mat(docfeats, train_ids)
        featmat_test, featurenames = features2mat(docfeats, test_ids, featurenames)
        # possibly normalize vectors
        if sim in ("cosine", "angularsim", "angulardist"):
            featmat_train /= np.linalg.norm(featmat_train, axis=0)
            featmat_test /= np.linalg.norm(featmat_test, axis=0)
        # compute kernel map
        K_map = featmat_test.dot(featmat_train.T).toarray()
        # transform further
        if sim in ("angularsim", "angulardist"):
            # make sure values are in the correct range (roundoff errors)
            K_map = np.minimum(K_map, 1)
            K_map = np.maximum(K_map, 0)
            if sim == "angularsim":
                K_map = 1. - 2. * np.arccos(K_map) / np.pi
            else:
                # distance
                K_map = 2. * np.arccos(K_map) / np.pi
    else:
        # compute similarity of all test examples to all training examples
        n_tr = len(train_ids)
        n_ts = len(test_ids)
        if not len(train_idx):
            train_idx = list(range(n_tr))
        K_map = np.array([[compute_sim(docfeats[did_ts], docfeats[train_ids[j]], sim)
                           for j in train_idx] for did_ts in test_ids])
    if normalize:
        train_sim = np.array([[compute_sim(docfeats[did_tr], docfeats[did_tr], sim) for did_tr in train_ids]])
        test_sim = np.array([[compute_sim(docfeats[did_ts], docfeats[did_ts], sim) for did_ts in test_ids]]).T
        norm = (np.tile(train_sim, (n_ts, 1)) + np.tile(test_sim, (1, n_tr))) / 2.
        K_map /= norm
    return K_map


def get_k_most_similar(K, row_ids, col_ids, did, k=10):
    """
    given the kernel matrix and corresponding docids, get the k most similar documents corresponding to the doc did

    Input:
        - K: kernel matrix or map with similarities (the higher the more similar)
        - row_ids: list of ids in the order of the rows in K
        - col_ids: list of ids in the order of the cols in K (same as row_ids in the case of the kernel matrix, not map)
        - did: a single docids corresponding to one row in K and to which similar documents should be found
        - k: how many similar documents should be returned
    Returns:
        - simdocids: [(docid, simscore)] a list of k docids and their respective similarity to the target did
    """
    try:
        dididx = row_ids.index(did)
    except ValueError:
        print("did must have a designated row in K")
        return []
    # get k indexes for K
    k_idx = np.flipud(np.argsort(K[dididx, :]))[:k + 1]
    return [(col_ids[i], K[dididx, i]) for i in k_idx if not col_ids[i] == did][:k]



from __future__ import unicode_literals, division, print_function, absolute_import
from builtins import range, str
import colorsys
import json
import numpy as np
import matplotlib.pyplot as plt
from .dict_utils import invert_dict0


def get_colors(N=100):
    HSV_tuples = [(x * 1. / (N+1), 1., 0.8) for x in range(N)]
    return [colorsys.hsv_to_rgb(*x) for x in HSV_tuples]


def colorindex(X, N=100):
    """
    X: list/vector for which different colors should be assigned
    N: number of colors that can be used
    """
    # get min & max values
    minv, maxv = min(X), max(X)
    # get a linear scale of the values
    lscale = np.linspace(minv, maxv, N)
    # for each value in X, find the index in the linspace
    return [np.nonzero(lscale >= val)[0][0] for val in X]


def pretty_coloring(X, varcol=0, N=100):
    """
    X: rows = observations, cols = variables
    varcol: variable for which different colors should be assigned
    N: number of colors that can be used
    """
    coloridx = colorindex(X[:, varcol], N)
    colors = np.array(get_colors(N))
    return colors[coloridx, :]


def prepare_viz(doc_ids, docdict, doccats, x, y, catdesc={}, filepath="docs.json", doc_ids_test=[], x_test=[], y_test=[]):
    """
    function to prepare text data for 2 dim visualization by saving a json file, that is a list of dicts,
    where each dict decodes 1 doc with "id" (doc_id), "x" and "y" (2dim coordinates derived from the kernel matrix
    using classical scaling), "title" (category/ies), "description" (whatever is in docdict at doc_id), "color" (for cat)
    Input:
        - doc_ids: list with keys for docdict and doccats
        - docdict: dict with docid:'description'
        - doccats: dict with docid: cat
        - x, y: 2d coordinates for all data points in the order of doc_ids (use x, y = proj2d(K, use_tsne, evcrit))
        - catdesc: category descriptions
        - filepath: where the json file will be saved
        - doc_ids_test, x_test, y_test: optional, same as before but for test points
    """
    # pretty preprocessing
    if not catdesc:
        categories = set(invert_dict0(doccats).keys())
        catdesc = {cat: cat for cat in categories}
    else:
        categories = list(catdesc.keys())
    colorlist = get_colors(len(categories))
    colordict = {cat: (255 * colorlist[i][0], 255 * colorlist[i][1], 255 * colorlist[i][2]) for i, cat in enumerate(sorted(categories))}
    # save as json
    print("saving json")
    data_json = []
    for i, key in enumerate(doc_ids):
        data_json.append({"id": key, "x": x[i], "y": y[i], "title": str(
            key) + " (%s)" % catdesc[doccats[key]], "description": docdict[key], "color": "rgb(%i,%i,%i)" % colordict[doccats[key]]})
    # if we have test points, do the same again
    for i, key in enumerate(doc_ids_test):
        data_json.append({"id": key, "x": x_test[i], "y": y_test[i], "title": str(key) + " (%s) - TEST POINT" % catdesc[
                         doccats[key]], "description": docdict[key], "color": "rgb(%i,%i,%i)" % colordict[doccats[key]]})
    with open(filepath, "w") as f:
        f.write(json.dumps(data_json, indent=2))


def basic_viz(doc_ids, doccats, x, y, catdesc={}, title="", doc_ids_test=[], x_test=[], y_test=[]):
    """
    plot a scatter plot of the data in 2d
    Input:
        - doc_ids: list with keys for docdict and doccats
        - doccats: dict with docid: cat
        - x, y: 2d coordinates for all data points in the order of doc_ids (use x, y = proj2d(K, use_tsne, evcrit))
        - catdesc: category descriptions (for legend)
        - doc_ids_test, x_test, y_test: optional, same as before but for test points (will get a higher alpha)
    """
    # pretty preprocessing
    if not catdesc:
        categories = set(invert_dict0(doccats).keys())
        catdesc = {cat: cat for cat in categories}
    else:
        categories = list(catdesc.keys())
    colorlist = get_colors(len(categories))
    colordict = {cat: (colorlist[i][0], colorlist[i][1], colorlist[i][2]) for i, cat in enumerate(sorted(categories))}
    # plot scatter plot
    plt.figure()
    for j, cat in enumerate(sorted(categories)):
        # get docids that belong to the current category
        didx_temp = [i for i, did in enumerate(doc_ids) if cat == doccats[did]]
        plt.plot(x[didx_temp], y[didx_temp], "o", label=catdesc[cat],
                 color=colordict[cat], alpha=0.6, markeredgewidth=0)
        # possibly do the same for test points
        if doc_ids_test:
            didx_temp = [i for i, did in enumerate(doc_ids_test) if cat == doccats[did]]
            plt.plot(x_test[didx_temp], y_test[didx_temp], "o", label=catdesc[
                     cat], color=colordict[cat], alpha=1., markeredgewidth=0)
    plt.xticks([], [])
    plt.yticks([], [])
    plt.xlim([x.min(), x.max()])
    plt.ylim([y.min(), y.max()])
    # plt.axis("equal")
    plt.title(title, fontsize=16)
    plt.legend(loc="center left", bbox_to_anchor=(1, 0.5), numpoints=1)
    # plt.tight_layout()


def json2plot(jsonpath, title="", baseline=False):
    data_json = json.load(open(jsonpath))
    doc_ids = [d["id"] for d in data_json]
    x = np.array([d["x"] for d in data_json])
    y = np.array([d["y"] for d in data_json])
    if baseline:
        catdesc = {"moviereviews neg": 1,
                   "moviereviews pos": 2,
                   "ohsumed C04": 3,
                   "ohsumed C10": 4,
                   "ohsumed C14": 5,
                   "ohsumed C23": 6,
                   "reuters coffee": 7,
                   "reuters crude": 8,
                   "reuters gold": 9,
                   "reuters grain": 10,
                   "reuters interest": 11,
                   "reuters money-supply": 12,
                   "reuters sugar": 13,
                   "reuters trade": 14,
                   "pmcpar abs-int-dis": 15,
                   "pmcpar methods": 16,
                   "pmcpar results": 17}
        doccats = {d["id"]: catdesc[d["title"].split("(")[-1][:-1]] for d in data_json}
        basic_viz(doc_ids, doccats, x, y, {v: k for k, v in catdesc.items()}, title)
    else:
        doccats = {d["id"]: d["title"].split("(")[-1][:-1] for d in data_json}
        basic_viz(doc_ids, doccats, x, y, {}, title)



from __future__ import unicode_literals, division, print_function, absolute_import
from collections import defaultdict
from sys import float_info as fi
from math import sqrt, log
import numpy as np


def _polynomial_sim(x, y, p=2):
    """
    polynomial/linear kernel
    compute the similarity between the dicts x and y with terms + their counts
    p: >1 for polynomial kernel (default = 2)
    """
    # get the words that occur in both x and y (for all others the product is 0 anyways)
    s = set(x.keys()) & set(y.keys())
    # nothing in common?
    if not s:
        return 0.
    return float(sum([x[word] * y[word] for word in s]))**p


def _sigmoidal_sim(x, y):
    """
    sigmoidal kernel
    compute the similarity between the dicts x and y with terms + their counts
    """
    # get the words that occur in both x and y (for all others the product is 0 anyways)
    s = set(x.keys()) & set(y.keys())
    # nothing in common?
    if not s:
        return 0.
    return np.tanh(sum([x[word] * y[word] for word in s]))


def _histint_sim(x, y):
    """
    histogram intersection kernel
    compute the similarity between the dicts x and y with terms + their counts
    """
    # get the words that occur in both x and y (for all others min is 0 anyways)
    s = set(x.keys()) & set(y.keys())
    # nothing in common?
    if not s:
        return 0.
    return float(sum([min(x[word], y[word]) for word in s]))


def _gaussian_sim(x, y, gamma=.25):
    """
    gaussian rbf kernel with width 1/gamma
    compute the similarity between the dicts x and y with terms + their counts
    """
    # get the words that occur in either x or y
    s = set(x.keys()) & set(y.keys())
    sx = set(x.keys()).difference(s)
    sy = set(y.keys()).difference(s)
    return np.exp(-gamma * float(sum([(x[word] - y[word])**2 for word in s]) + sum([(x[word])**2 for word in sx]) + sum([(y[word])**2 for word in sy])))


def _minkowski_sim(x, y, p=3):
    """
    minkowski distance: p=1: manhattan, p=2: squared euclidean, etc. returned with - as a pseudo similarity
    compute the similarity between the dicts x and y with terms + their counts
    """
    # get the words that occur in either x or y
    s = set(x.keys()) & set(y.keys())
    sx = set(x.keys()).difference(s)
    sy = set(y.keys()).difference(s)
    return 1. / (fi.epsilon + float(sum([np.abs(x[word] - y[word])**p for word in s]) + sum([x[word]**p for word in sx]) + sum([y[word]**p for word in sy])))


def _canberra_sim(x, y, p=1):
    """
    canberra distance p=1 / chi^2 with p=2; returned with - as a pseudo similarity
    compute the similarity between the dicts x and y with terms + their counts
    """
    # get the words that occur in either x or y
    s = set(x.keys()) | set(y.keys())
    x = defaultdict(float, x)
    y = defaultdict(float, y)
    return 1. / (fi.epsilon + float(sum([np.abs(x[word] - y[word])**p / float(x[word] + y[word]) for word in s])))


def _chebyshev_sim(x, y):
    """
    1 - chebyshev distance, as a pseudo similarity, only works if counts are normalized to be <= 1
    compute the similarity between the dicts x and y with terms + their counts
    """
    # get the words that occur in either x or y
    s = set(x.keys()) | set(y.keys())
    x = defaultdict(float, x)
    y = defaultdict(float, y)
    return 1. - float(max([np.abs(x[word] - y[word]) for word in s]))


def _hellinger_sim(x, y):
    """
    hellinger distance, returned with - as a pseudo similarity
    compute the similarity between the dicts x and y with terms + their counts
    """
    # get the words that occur in either x or y
    s = set(x.keys()) & set(y.keys())
    sx = set(x.keys()).difference(s)
    sy = set(y.keys()).difference(s)
    return 1. / (fi.epsilon + (sum([(sqrt(float(x[word])) - sqrt(float(y[word])))**2 for word in s]) + float(sum([x[word] for word in sx])) + float(sum([y[word] for word in sy]))))


def _jenshan_sim(x, y):
    """
    jensen-shannon distance, returned with - as a pseudo similarity
    compute the similarity between the dicts x and y with terms + their counts
    """
    # get the words that occur in either x or y
    s = set(x.keys()) | set(y.keys())
    x = defaultdict(float, x)
    y = defaultdict(float, y)

    def H(xcount, ycount):
        h = 0.
        if xcount:
            h += xcount * log(2. * xcount / (xcount + ycount))
        if ycount:
            h += ycount * log(2. * ycount / (xcount + ycount))
        return h
    return 1. / (fi.epsilon + float(sum([H(float(x[word]), float(y[word])) for word in s])))


def _simpson_sim(x, y):
    """
    simpson similarity
    compute the similarity between the dicts x and y with terms + their counts
    """
    return _histint_sim(x, y) / min(_histint_sim(x, x), _histint_sim(y, y))


def _braun_sim(x, y):
    """
    braun-blanquet similarity
    compute the similarity between the dicts x and y with terms + their counts
    """
    return _histint_sim(x, y) / max(_histint_sim(x, x), _histint_sim(y, y))


def _kulczynski_sim(x, y):
    """
    kulczynski (2) similarity
    compute the similarity between the dicts x and y with terms + their counts
    """
    a = _histint_sim(x, y)
    return 0.5 * (a / _histint_sim(x, x) + a / _histint_sim(y, y))


def _jaccard_sim(x, y):
    """
    jaccard similarity
    compute the similarity between the dicts x and y with terms + their counts
    """
    a = _histint_sim(x, y)
    return a / (_histint_sim(x, x) + _histint_sim(y, y) - a)


def _dice_sim(x, y):
    """
    czekanowski, sorensen-dice similarity
    compute the similarity between the dicts x and y with terms + their counts
    """
    return 2. * _histint_sim(x, y) / (_histint_sim(x, x) + _histint_sim(y, y))


def _otsuka_sim(x, y):
    """
    otsuka, ochiai similarity
    compute the similarity between the dicts x and y with terms + their counts
    """
    return _histint_sim(x, y) / (sqrt(_histint_sim(x, x)) * sqrt(_histint_sim(y, y)))


def _sokal_sim(x, y):
    """
    sokal-sneath, anderberg similarity
    compute the similarity between the dicts x and y with terms + their counts
    """
    a = _histint_sim(x, y)
    return a / (2. * _histint_sim(x, x) + 2. * _histint_sim(y, y) - 3. * a)


def compute_sim(x, y, sim="linear"):
    """
    compute the similarity between x and y
    Input:
        - x, y: dicts representing 2 documents (all normalization, weighting, etc has to be done before!)
        - sim: the kind of similarity to be computed:
                -- kernels: ['linear', 'polynomial', 'sigmoidal', 'histint', 'gaussian']
                -- (neg) distance: ['manhattan', 'sqeucl', 'minkowski', 'canberra', 'chisq', 'chebyshev', 'hellinger', 'jenshan']
                -- similarity coef: ['simpson', 'braun', 'kulczynski', 'jaccard', 'dice', 'otsuka', 'sokal']
    Returns:
        - a single number representing the similarity between x and y
    """
    s = None
    try:
        if sim == "linear":
            s = _polynomial_sim(x, y, 1)
        elif sim == "polynomial":
            s = _polynomial_sim(x, y)
        elif sim == "sigmoidal":
            s = _sigmoidal_sim(x, y)
        elif sim == "histint":
            s = _histint_sim(x, y)
        elif sim == "gaussian":
            s = _gaussian_sim(x, y)
        elif sim == "minkowski":
            s = _minkowski_sim(x, y)
        elif sim == "sqeucl":
            s = _minkowski_sim(x, y, 2)
        elif sim == "manhattan":
            s = _minkowski_sim(x, y, 1)
        elif sim == "canberra":
            s = _canberra_sim(x, y)
        elif sim == "chisq":
            s = _canberra_sim(x, y, 2)
        elif sim == "chebyshev":
            s = _chebyshev_sim(x, y)
        elif sim == "hellinger":
            s = _hellinger_sim(x, y)
        elif sim == "jenshan":
            s = _jenshan_sim(x, y)
        elif sim == "simpson":
            s = _simpson_sim(x, y)
        elif sim == "braun":
            s = _braun_sim(x, y)
        elif sim == "kulczynski":
            s = _kulczynski_sim(x, y)
        elif sim == "jaccard":
            s = _jaccard_sim(x, y)
        elif sim == "dice":
            s = _dice_sim(x, y)
        elif sim == "otsuka":
            s = _otsuka_sim(x, y)
        elif sim == "sokal":
            s = _sokal_sim(x, y)
        else:
            print("ERROR: sim not known!!")
    except ZeroDivisionError:
        s = 0.
    return s



from __future__ import unicode_literals, division, print_function, absolute_import
import numpy as np
from copy import deepcopy


def invert_dict0(adict):
    """
    for a key:value dict, return value:[key(s)],
    e.g. dict[doc] = label --> dict[label] = [docs]
    """
    inv_dict = {}
    [inv_dict.setdefault(v, []).append(k) for k, v in adict.items()]
    return inv_dict


def invert_dict1(adict):
    """
    for a key:[value(s)] dict, return value:[key(s)],
    e.g. dict[doc] = [terms] --> dict[term] = [docs]
    """
    inv_dict = {}
    [inv_dict.setdefault(v, []).append(k) for k, vlist in adict.items() for v in vlist]
    return inv_dict


def invert_dict2(adict):
    """
    for a dict(key:dict(key2:value)), return dict(key2:dict(key:value)),
    e.g. dict[doc] = dict[term]:count --> dict[term] = dict[doc]:count
    """
    inv_dict = {}
    [inv_dict.setdefault(key2, {}).update({key: v}) for key, dict2 in adict.items()
     for key2, v in dict2.items()]
    return inv_dict


def select_copy(adict, key_ids):
    """
    make a copy of adict but including only the keys in key_ids
    """
    adict_copy = {}
    for key in key_ids:
        try:
            adict_copy[key] = deepcopy(adict[key])
        # ignore keys that are not in the original dict
        except KeyError:
            pass
    return adict_copy


def vec2dict(vector):
    """
    Input:
        vector: a list or (1d) numpy array
    Returns:
        a dict with {idx:value} for all values of the vector
    """
    return {i: val for i, val in enumerate(vector)}


def norm_dict(somedict, norm="max", n_all=0):
    """
    Input:
        somedict: a dictionary with values of something
        norm (either 'max', 'sum', or 'length'): how to normalize the values
        n_all: for 'mean' and 'std' we need to know the total number of values
               (if the dict is sparse, zeros are not included)
    Output:
        a dictionary with the normalized values
    """
    # can't normalize empty dicts
    if not somedict:
        return somedict
    if norm == "binary":
        return {k: 1. for k in somedict}
    elif norm == "sum":
        N = float(sum(somedict.values()))
    elif norm == "max":
        N = float(max(np.abs(list(somedict.values()))))
    elif norm == "length":
        N = np.linalg.norm(list(somedict.values()))
    elif norm == "mean":
        N = np.mean(list(somedict.values()) + [0.] * (n_all - len(somedict)))
    elif norm == "std":
        N = np.std(list(somedict.values()) + [0.] * (n_all - len(somedict)))
    else:
        print("ERROR: norm not known!!")
        return somedict
    if N:
        return {s: somedict[s] / N for s in somedict}
    else:
        return somedict


def combine_dicts(a, b, op=max):
    """
    Input:
        a, b: dictionaries with values of something
        op: how the values should be combined (max or sum (pass the function, not a string!))
    Output:
        a dictionary with all values from both dictionary. if a value occurred in only one of the
            dicts it is returned as it was, otherwise the values corresponding to a key from both
            dicts are combined according to op (e.g. add values together or get max of all values)
    """
    return dict(list(a.items()) + list(b.items()) + [(k, op([a[k], b[k]])) for k in set(b) & set(a)])



from __future__ import unicode_literals, division, print_function, absolute_import
from builtins import range
import numpy as np


def xval(Xlist, K=10):
    """
    Input:
        - Xlist: list of ids [tip: maybe use np.random.permutation(len(Xlist)) before]
        - K: number of folds
    Generator yields:
        - training and test sets with ids
    """
    for k in range(K):
        train = [x for i, x in enumerate(Xlist) if i % K != k]
        test = [x for i, x in enumerate(Xlist) if i % K == k]
        yield train, test


def balanced_xval(Xdict, K=10, random_seed=None):
    """
    Input:
        - Xdict: dictionary with category : list of doc_ids
        - K: number of folds
    Generator yields:
        - training and test sets that are balanced with respect to
          the number of docs from each category (i.e. ratio stays the same)
    """
    if random_seed is not None:
        np.random.seed(random_seed)
    Xlist = []
    for cat in Xdict:
        Xlist.extend(list(np.random.permutation(sorted(Xdict[cat]))))
    for k in range(K):
        train = [x for i, x in enumerate(Xlist) if i % K != k]
        test = [x for i, x in enumerate(Xlist) if i % K == k]
        yield train, test



from __future__ import unicode_literals, division, print_function, absolute_import
import numpy as np
from .dict_utils import invert_dict1, select_copy


def knn(K_map, train_ids, test_ids, doccats, k=25, adapt=True, alpha=5, weight=True):
    """
    k nearest neighbors
    Input:
        - K_map: matrix of size len(test_ids)xlen(train_ids) with similarities of the test to the training docs
        - train_ids: list of document ids in the order used for K_map
        - test_ids: list of document ids that need to be assigned a category
        - doccats: true categories of training documents (as a list, since multiple categories per document are allowed)
        - k (default=25): how many nearest neighbors of 1 category should be considered (at most)
        - adapt (default=True): if the k value should be adapted (for skewed category distributions to avoid the bias of
                categories with many samples)
        - alpha (default=5): if adapt=True, how many samples should be considered at least
        - weight (default=True): if the nearest neighbors should be weighted by their similarity
    Returns:
        - for every test document a likeliness score for every category: dict[tid] = dict[cat]:score (between 0 and 1)
    """
    categories = sorted(invert_dict1(doccats).keys())
    # select from doccats only training examples
    doccats = select_copy(doccats, train_ids)
    # dict with cat:[docs]
    cat_docs = invert_dict1(doccats)
    for cat in categories:
        if cat not in cat_docs:
            cat_docs[cat] = []
    # k for every category
    if adapt:
        # max number of samples in category
        cat_max = max([len(cat_docs[cat]) for cat in categories])
        # adaptive k is at least alpha and at most k or number of samples in category
        k_cat = {cat: min(len(cat_docs[cat]), max(int(alpha), int(k * len(cat_docs[cat]) / cat_max))) for cat in categories}
    else:
        k_cat = {cat: k for cat in categories}
    # get index for k nearest neighbors
    k_idx = np.fliplr(np.argsort(K_map))[:, :k]
    train_ids_dict = {doc: i for i, doc in enumerate(train_ids)}
    likely_cat = {did_ts: {} for did_ts in test_ids}
    for cat in categories:
        # get K_map index of all training documents that belong to the category
        cdoc_idx = set([train_ids_dict[doc] for doc in cat_docs[cat]])
        # compute the score for that category for every test example
        for i, did_ts in enumerate(test_ids):
            # get overlap between category specific training examples and the
            # (adapted) k nearest neighbors of the test example
            tidx = sorted(set(cdoc_idx & set(k_idx[i, :k_cat[cat]])))
            if tidx and np.sum(K_map[i, tidx]):
                # get score
                if weight:
                    # sum of similarity of k nearest neighbors of category cat / sum of similarity of all knn
                    likely_cat[did_ts][cat] = np.sum(K_map[i, tidx]) / np.sum(K_map[i, k_idx[i, :k_cat[cat]]])
                else:
                    # number of nearest neighbors of category cat
                    likely_cat[did_ts][cat] = float(len(tidx)) / k_cat[cat]
            else:
                likely_cat[did_ts][cat] = 0.
    return likely_cat


def get_labels(likely_cat, threshold="max"):
    """
    Input:
        - likely_cat: the confidence scores for every category for every test doc as a dict (scores normalized to 1)
        - threshold: threshold for the likeliness score: selects as many categories as have a score equal to or above the
                   threshold (between 0 and 1). If threshold is set to 'max', only the category with the highest score
                   is chosen (unless all categories have a score of 0)
    Returns:
        - labels: dict with doc:[cats] for each test doc where the list contains as many categories as have scores above threshold
                 --> the categories are chosen as the highest scoring categories in likely_cat
    Note:
        if threshold='max' and all categories have score of 0, a random category is chosen,
        otherwise if threshold is a float and no category has a score above threshold, the list will be empty
    """
    labels = {}
    for tid in likely_cat:
        # either take the most likely category
        if threshold == "max":
            labels[tid] = [max(likely_cat[tid].keys(), key=likely_cat[tid].get)]
        # or all categories with a score above threshold
        else:
            labels[tid] = [cat for cat in likely_cat[tid] if likely_cat[tid][cat] >= threshold]
    return labels



from __future__ import unicode_literals, division, print_function, absolute_import
from builtins import range
import numpy as np
from scipy.sparse.linalg import eigsh
from scipy.spatial.distance import pdist, squareform


def tsne_sim(S, no_dims=2, earlystop=True, init="random", verbose=True, max_iter=1500):
    """
    TSNE_Sim Performs symmetric t-SNE on similarity matrix S
       mappedX = tsne_sim(S, no_dims)

     The function performs symmetric t-SNE on pairwise similarity matrix S
     to create a low-dimensional map of no_dims dimensions (default = 2).
     The matrix S is assumed to be symmetric, sum up to 1, and have zeros
     on the diagonal.
     The low-dimensional data representation is returned in mappedX.
    """
    # Initialize some variables
    n = S.shape[0]                                     # number of instances
    momentum = 0.5                                     # initial momentum
    final_momentum = 0.8                               # value to which momentum is changed
    mom_switch_iter = 250                              # iteration at which momentum is changed
    exageration = 8.                                   # by how much we lie about the P-values
    stop_lying_iter = 100                              # iteration at which lying about P-values is stopped
    epsilon = 500.                                     # initial learning rate
    min_gain = .01                                     # minimum gain for delta-bar-delta
    # initialize the solution
    if init == "kpca" and no_dims == 2:
        # kPCA initialization
        x, y = proj2d(S, use_tsne=False)
        mappedX = np.zeros((n, no_dims))
        mappedX[:, 0] = x
        mappedX[:, 1] = y
    else:
        mappedX = .0001 * np.random.randn(n, no_dims)
    y_incs = np.zeros(mappedX.shape)
    gains = np.ones(mappedX.shape)
    last_cost = np.inf
    # Make sure S-vals are set properly
    np.fill_diagonal(S, 0.)                            # set diagonal to zero
    S = 0.5 * (S + S.T)                                # symmetrize S-values
    S /= np.sum(S)                                     # make sure S-values sum to one
    S = np.maximum(S, 1e-12)
    const = np.sum(S * np.log(S))                      # constant in KL divergence
    S *= exageration                                   # lie about the S-vals to find better local minima
    # Run the iterations
    for itr in range(max_iter):
        # Compute joint probability that point i and j are neighbors
        # sum_mappedX = np.sum(np.square(mappedX), 1)
        # Student-t distribution
        # num = 1 / (1 + np.add(np.add(-2 * np.dot(mappedX, mappedX.T), sum_mappedX).T, sum_mappedX))
        num = 1 / (1 + squareform(pdist(mappedX, "sqeuclidean")))
        np.fill_diagonal(num, 0.)                       # set diagonal to zero
        Q = num / np.sum(num)                             # normalize to get probabilities
        Q = np.maximum(Q, 1e-12)
        # Compute the gradients (faster implementation)
        L = (S - Q) * num
        # free some memory
        del num
        y_grads = 4. * np.dot((np.diag(np.sum(L, 0)) - L), mappedX)
        # Update the solution (note that the y_grads are actually -y_grads)
        gains = (gains + .2) * np.invert(np.sign(y_grads) == np.sign(y_incs)) + \
            (gains * .8) * (np.sign(y_grads) == np.sign(y_incs))
        gains[gains < min_gain] = min_gain
        y_incs = momentum * y_incs - epsilon * (gains * y_grads)
        mappedX += y_incs
        mappedX -= np.tile(np.mean(mappedX, 0), (n, 1))
        # Update the momentum if necessary
        if itr == mom_switch_iter:
            momentum = final_momentum
        if itr == stop_lying_iter:
            S /= exageration
        # Print out progress
        if not (itr + 1) % 25:
            if itr < stop_lying_iter:
                cost = const - np.sum(S / exageration * np.log(Q))
            else:
                cost = const - np.sum(S * np.log(Q))
                if earlystop and itr > mom_switch_iter and cost >= last_cost - 0.000001:
                    break
                else:
                    last_cost = cost
            if verbose:
                print("Iteration %i: error is %.5f" % (itr + 1, cost))
    return mappedX


def classical_scaling(K, nev=2, evcrit="LM"):
    """
    compute 2 eigenvalues and -vectors of a similarity matrix
    Input:
        - K: a symmetric nxn similarity matrix (for dissimilarity matrices, first multiply by -1/2)
        - nev: how many eigenvalues should be computed
        - evcrit: how eigenvalues should be selected ('LM' for largest real part, 'SM' for smallest real part)
    Return:
        - mappedX: n x 2 matrix of mapped data
    """
    n, m = K.shape
    H = np.eye(n) - np.tile(1. / n, (n, n))
    B = np.dot(np.dot(H, K), H)
    D, V = eigsh((B + B.T) / 2., k=nev, which=evcrit)  # guard against spurious complex e-vals from roundoff
    return np.dot(V.real, np.diag(np.sqrt(np.abs(D.real))))


def proj2d(K, use_tsne=True, evcrit="LM", max_iter=1500, verbose=True):
    """
    wrapper function to project data to 2D
    """
    if use_tsne:
        if verbose:
            print("performing tSNE: %i datapoints" % K.shape[0])
        X = tsne_sim(K, max_iter=max_iter, verbose=verbose)
        x = X[:, 0]
        y = X[:, 1]
    else:
        if verbose:
            print("performing classical scaling: %i datapoints" % K.shape[0])
        if evcrit == "LM" or evcrit == "SM":
            X = classical_scaling(K, evcrit=evcrit)
            x = X[:, 0]
            y = X[:, 1]
        elif evcrit == "LS":
            x = classical_scaling(K, nev=1, evcrit="LM")
            y = classical_scaling(K, nev=1, evcrit="SM")
        else:
            print("ERROR: evcrit not known!")
            return None
    return x, y
